/***********************************************/
/* Parser para o protocolo HTTP                */
/*                                             */
/* Eleri Cardozo, Fev 2003                     */
/* Rev. Fev 2007: trata LF                     */
/* Rev. Jul 2008: destrutor nas struts (fpinho)*/
/***********************************************/

#include "http_parser.h"
#include <time.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>


/* construtores e destrutores */
http_parametro::http_parametro()
{
    nome = NULL;
    valores = NULL;
    proximo = NULL;
    anterior = NULL;
    nvalores = 0;
}

http_parametro::~http_parametro()
{
  for(int i = 0; i < nvalores; i++) 
    {
      free(valores[i]);
    }
  free(valores);
  free(nome);
}

http_cabecalho::http_cabecalho()
{
  operacao = NULL;
  recurso = NULL; 
  versao = NULL;  
  primeiro = NULL;
  ultimo = NULL;    
}

http_cabecalho::~http_cabecalho()
{
    if (operacao != NULL) free(operacao);
    if (recurso != NULL) free(recurso);
    if (versao != NULL) free(versao);
    
    http_parametro *http_next, *http_current = primeiro;
    while(http_current) 
      {
        http_next = http_current->proximo;
        delete http_current;
        http_current = http_next;
      }
}

linha_cab::linha_cab()
{
  linha = NULL;
  tamanho = 0;
}

linha_cab::~linha_cab()
{
  if (linha != NULL)
    {
      free(linha);
    }
}

/* funcao que le uma linha do cabecalho */
/* separa linha por LF ou CR-LF */
/* retorna linha vazia se encontrar duas linhas em branco (fim do cabecalho) */
/* ou NULL em caso de erro */

linha_cab* read_linha_cab(int file) {
    char buff[256];    /* buffer auxiliar */
    linha_cab *hl;     /* retorno */
    int i = 0;
    int err;

    do {
        err = read(file, &buff[i], 1);
        if(err == -1) return(NULL);  /* erro na chamada read */
        if(err == 1)   /* lido 1 byte */
        {
            if(buff[i] == 13) continue;    /* elimina CR */
            if(i == 0 && buff[i] == 10)    /* linha em branco -- fim de cabecalho */
            {
                hl = new linha_cab;
                hl->tamanho = 0;
                hl->linha = (char *)malloc(1);
                *(hl->linha) = '\0';
                return(hl);
            }

            if(buff[i] != 10)  /* caractere que nao LF */
            {
                i++;
                continue;
            }
            else   /* fim da linha */
            {
                hl = new linha_cab;
                hl->tamanho = i;
                hl->linha = (char *)malloc(i+1);
                strncpy(hl->linha, buff, i);
                *(hl->linha + i) = '\0';
                return(hl);
            }
        }
        /* err == 0: cliente fechou a conexao */
        printf("Connection closed by peer.\n");
        fflush(stdout);
        return(NULL);
    } while(i < 256);
    printf("Linha muito longa.\n");
    return(NULL);
}



/* funcao que processa a primeira linha do cabecalho */
/* preenche uma estrutura http_cabecalho passada como parametro */
/* retorna 0 (OK) ou -1 (erro) */

int primeira_linha(http_cabecalho *h, char *linha) {
  char *op, *res, *ver, *s;
    int i;
    /* obtem operacao */
    while(*linha && *linha == ' ') linha++; /* remove espaco inicial caso exista  */
    if(! *linha) return -1;
    op = linha;
    i = 0;
    while(*linha && *linha != ' ') {linha++; i++;}  /* caminha ate prox. espaco */
    if(i == 0) return(-1);
    h->operacao = (char *)malloc(i + 1);
    strncpy(h->operacao, op, i);
    *(h->operacao + i) = '\0';
    if(strcmp(h->operacao, "GET") != 0 &&
            strcmp(h->operacao, "POST") != 0 &&
            strcmp(h->operacao, "OPTIONS") != 0 &&
            strcmp(h->operacao, "TRACE") != 0) return(-1);

    /* obtem o recurso */
    while(*linha && *linha == ' ') linha++;   /* remove espaco inicial */
    res = linha;
    i = 0;
    while(* linha && *linha != ' ') {linha++; i++;}  /* caminha ate prox. espaco */
    if(i == 0) return(-1);
    h->recurso = (char *)malloc(i + 1);
    strncpy(h->recurso, res, i);
    *(h->recurso + i) = '\0';
    s = h->recurso;
    while(*s && *s != '?') s++;
    if(*s) {
      *s = '\0';
      h->cgi_pars = ++s;
    }
    else h->cgi_pars = NULL;

    /* obtem versao do protocolo */
    while(*linha && *linha == ' ') linha++;   /* remove espaco inicial */
    ver = linha;
    i = 0;
    while(*linha && *linha != ' ') {linha++; i++;}  /* caminha ate prox. espaco */
    if(i == 0) return(-1);
    h->versao = (char *)malloc(i + 1);
    strncpy(h->versao, ver, i);
    *(h->versao + i) = '\0';

    return(0);
}



/* funcao que processa parametros do cabecalho  */
/* adiciona parametros a uma estrutura http_cabecalho */ 
/* retorna 0 (OK) ou -1 (erro) */

int http_par(http_cabecalho *hddr, char *linha) {
    http_parametro *par;
    int i, k;
    char *nome, *str;
    while(*linha && *linha == ' ') linha++; /* remove espaco inicial caso exista  */
    nome = linha;
    i = 0;
    while(*linha && *linha != ':') {linha++; i++;}  /* posiciona em : */
    if(i == 0) return(-1);
    /* aloca celula da lista ligada */
    par = (http_parametro *)malloc(sizeof(http_parametro));
    par->proximo = NULL;
    par->anterior = NULL;
    par->nvalores = 0;
    par->nome = (char *)malloc(i + 1);
    strncpy(par->nome, nome, i);
    *(par->nome + i) = '\0';

    /* conta numero de valores associado ao parametro */
    linha++;
    str = linha;
    while(*str) {if(*str == ',') par->nvalores++; str++;}
    par->nvalores++;

    /* aloca ponteiro para valores */
    par->valores = (char **)malloc(par->nvalores * sizeof(char*));

    /* busca e armazena  valores */
    for(k = 0; k < par->nvalores; k++) {
        while(*linha && *linha == ' ') linha++;   /* remove espaco, caso exista  */
        str = linha;
        i = 0;
        while(*linha && *linha != ',') {linha++; i++;}  /* posiciona em , */
        if(*linha) linha++;
        par->valores[k] = (char *)malloc(i + 1);
        strncpy(par->valores[k], str, i);
        *(par->valores[k] + i) = '\0';

    }

    /* liga parametro na lista de parametros */
    if(hddr->primeiro == NULL) {   /* lista vazia */
        hddr->primeiro = par;
        hddr->ultimo = par;
    }
    else {
        par->anterior = hddr->ultimo;
        hddr->ultimo->proximo = par;
        hddr->ultimo = par;
    }

    return(0);
}


/* funcao que imprime o cabecalho processado */

void imprime_cabecalho(http_cabecalho *hddr, int file_out) {
    int i;
    http_parametro *par;
    write(file_out, hddr->operacao, strlen(hddr->operacao));
    write(file_out, " ", 1);
    write(file_out, hddr->recurso, strlen(hddr->recurso));
    if(hddr->cgi_pars) {
      write(file_out, "?", 1);
      write(file_out, hddr->cgi_pars, strlen(hddr->cgi_pars));
    }
    write(file_out, " ", 1);
    write(file_out, hddr->versao, strlen(hddr->versao));
    write(file_out, "\n", 1);
    par = hddr->primeiro;
    while(par) {
        write(file_out, par->nome, strlen(par->nome));
        write(file_out, ": ", 2);
        for(i = 0; i < par->nvalores; i++) {
            write(file_out, par->valores[i], strlen(par->valores[i]));
            if(par->nvalores > 1 && i < par->nvalores - 1) write(file_out, ", ", 2);
        }
        write(file_out, "\n", 1);
        par = par->proximo;
    }
}



/* programa teste: requer arquivo de entrada e arquivo para saida
   como argumentos do programa */

main(int argc, char** argv)
{
  int arq_in = 0;       /* arquivos de entrada/sada - stdin default */
  int arq_out = 1;      /* stdout default */
  http_cabecalho *hddr; /* estrutura do cabecalho */
  linha_cab *hl;        /* cabecalho http */
  int numlinha = 0;     /* linhas do cabecalho */

  if(argc > 1) {    /* entrada */
    if((arq_in = open(argv[1], O_RDONLY)) <= 0) {
      printf("%s: arquivo inexistente\n", argv[1]);
      exit(0);
    }
  }

  if(argc > 2) {    /* saida */
      if((arq_out = open(argv[2], O_RDWR | O_CREAT, 0655)) <= 0) {
        printf("%s: nao pode ser criado\n, argv[2]");
        exit(0);
      }
    }


  /* aloca cabecalho */
  hddr =  new http_cabecalho();

  /* le linha por linha ate fim de arquivo */
  numlinha = 0;  /* zere sempre que iniciar o parsing */
  while((hl = read_linha_cab(arq_in)) != NULL)
    {
      if(hl->tamanho == 0) {   /* fim do cabecalho */
	delete hl;   /* librea linha lida */
	break;
      }
      if(numlinha == 0) {   /* primeira linha */
	if(primeira_linha(hddr, hl->linha) == -1) {
	  printf("Erro de sintaxe.\n");
	  exit(0);
	}
      }
      else {   /* linhas subsequentes*/
	if(http_par(hddr, hl->linha) == -1) {
	  printf("Erro de sintaxe.\n");
	  exit(0);
	}
      }
      delete hl;   /* librea linha lida */
      numlinha++;
    }

  /* imprime cabecalho */
  imprime_cabecalho(hddr, arq_out);
  delete hddr;  /* libera cabecalho */
}
